/*
	Commands by Catid@pacbell.net
*/

#ifndef COMMAND_H
#define COMMAND_H

#include "datatypes.h"
#include "math.h"
#include "player.h"
#include <stdio.h>

class Host;

struct _switch
{					//            |----|
	char type;		// !command -a_=blah-s blah
	char * param;	// !command -at=____-s blah

	_switch(char t, char * p)
	{
		Uint32 len = 0;
		while (p[len++]);	// len == strlen(msg) + 1

		param = new char[len];
		if (param) MEMCPY(p, param, len);

		type = t;
	}

	~_switch()
	{
		if (param) delete []param;
	}
};

struct Command
{
	_linkedlist <_switch> slist;

	char * cmd;		// !_______ -at=blah-s blah
	char * final;	// !command -at=blah-s ____
	Host * host;

	Command(char * msg, Host * h)
	{
		host = h;

		Uint32 len = 0;
		while (msg[len++]);	// len == strlen(msg) + 1
		char * tmp = new char[len];
		Uint32 index = 0;

		cmd = new char[len];
		cmd[0] = '\0';

		final = new char[len];
		final[0] = '\0';

		bool inStub  = true,
			 inParam = false,
			 inFinal = false,
			 seenSpace = false;

		for (Uint32 i = 0; i < len; i++)
		{
			char c = msg[i];

			switch (c)
			{
			case ' ':
				{
					if (inStub)
					{
						tmp[index++] = 0;

						MEMCPY(tmp, cmd, index);

						inStub = false;

						index = 0;
					}
					else if (inParam)
					{
						tmp[index++] = 0;

						addParam(tmp);

						inParam = false;

						index = 0;
					}
					else if (inFinal)
					{
						tmp[index++] = c;
					}

					seenSpace = true;
				}
				break;
			case '-':
				{
					if (inStub)
					{
						tmp[index++] = 0;

						MEMCPY(tmp, cmd, index);

						inStub = false;
						inParam = true;

						index = 0;
					}
					else if (inParam)
					{
						tmp[index++] = 0;

						addParam(tmp);

						index = 0;
					}
					else if (inFinal)
					{
						tmp[index++] = c;
					}
					else
					{
						inParam = true;
					}

					seenSpace = false;
				}
				break;
			case '\0':
				{
					tmp[index++] = 0;

					if (inStub)
					{
						MEMCPY(tmp, cmd, index);
					}
					else if (inParam)
					{
						addParam(tmp);
					}
					else
					{
						MEMCPY(tmp, final, index);
					}

					seenSpace = false;

					delete []tmp;
				}
				return;
			default:
				if (inStub)
				{
					c = tolower(c);
				}
				else if (seenSpace)
				{
					inFinal = true;

					seenSpace = false;
				}

				tmp[index++] = c;
			};
		}
	}

	bool check(char * msg)
	{
		Uint32 len = 0;
		while (msg[len++]);	// len == strlen(msg) + 1

		for (Uint32 i = 0; i < len; i++)
		{
			if (msg[i] != cmd[i])
				return false;
		}

		return true;
	}

	bool checkParam(char * msg)
	{
		Uint32 len = 0;
		while (msg[len++]);	// len == strlen(msg) + 1

		for (Uint32 i = 0; i < len; i++)
		{
			if (msg[i] != final[i])
				return false;
		}

		return true;
	}

	_switch * getParam(char type)
	{
		_listnode <_switch> * parse = slist.head;

		while (parse)
		{
			_switch * s = parse->item;

			if (s->type == type)
				return s;

			parse = parse->next;
		}

		return NULL;
	}

	void addParam(char * msg)
	{
		Uint32 index = 0,
			   len   = 0;
		bool inParams = true;
		while (msg[len++]);	// len == strlen(msg) + 1

		char * tmp = new char[len];
		char type = '\0';

		for (Uint32 i = 0; i < len; i++)
		{
			char c = msg[i];

			switch (c)
			{
			case '=':
				{
					inParams = false;
				}
				break;
			case '\0':
				{
					if (type && (inParams == false))
					{
						tmp[index++] = 0;

						slist.kill(slist.tail, false);
						_switch * s = new _switch(type, tmp);
						if (s) slist.append(s);
					}
				}
				return;
			default:
				if (inParams)
				{
					type = tolower(c);
					_switch * s = new _switch(type, "");
					if (s) slist.append(s);
				}
				else
				{
					tmp[index++] = c;
				}
			};
		}
	}

	~Command()
	{
		delete []cmd;
		delete []final;
	}
};

void gotCommand(Host * h, Player * p, char * m);

#endif	// COMMAND_H
