#include "encrypt.h"
#include "checksum.h"
#include "math.h"


//////// SubSpace packet encryption ////////
// Courtesy of Coconut Emulator

SS_ENCR::SS_ENCR()
{
	key = 0;
	sentKey = 0;
}

Uint32 SS_ENCR::generateKey()
{
	if (sentKey)
		return sentKey;

	DWORD edx = GetTickCount() * 0xCCCCCCCD;

	long res = (keygen.getNext() << 16) + (edx >> 3) + keygen.getNext();

	res = (res ^ edx) - edx;

	if (res > 0)
		res = -res;

	return (sentKey = res);
}

Uint32 SS_ENCR::getSessionKey(Uint32 clientKey)
{
	return ((~clientKey) + 1);	// Two's complement
}

bool SS_ENCR::validateSessionKey(Uint32 key)
{
	return ((key == sentKey) || (key == getSessionKey(sentKey)));
}

void SS_ENCR::initializeEncryption(Uint32 key)
{
	if (sentKey == key)
	{
		this->key = 0;
	}
	else
	{
		this->key = key;

		Uint16 * stream = (Uint16*)keystream;

		prng.seed(key);

		for (int i = 0; i < (520 / sizeof(Uint16)); i++)
		{
			stream[i] = prng.getNextE();
		}
	}
}

void SS_ENCR::encrypt(char * msg, Uint32 len)
{
	if (key == 0) return;

	long IV = key;

	char buffer[520];
	memset(buffer, 0, 520);

	if ((msg[0] == 0) && (len > 1))
		MEMCPY(msg + 2, buffer, len - 2);
	else
		MEMCPY(msg + 1, buffer, len - 1);

	for (short Counter = 0; Counter < 520; Counter += 4)
	{
		IV = getLong(buffer, Counter) ^ getLong(keystream, Counter) ^ IV;

		*(long*)&buffer[Counter] = IV;
	}

	if (msg[0] == 0)
		MEMCPY(buffer, msg + 2, len - 2);
	else
		MEMCPY(buffer, msg + 1, len - 1);
}

void SS_ENCR::decrypt(char * msg, Uint32 len)
{
	if (key == 0) return;

	long IV = key;
	long EDX;

	char buffer[520];
	memset(buffer, 0, 520);

	if ((msg[0] == 0) && (len > 1))
		MEMCPY(msg + 2, buffer, len - 2);
	else
		MEMCPY(msg + 1, buffer, len - 1);

	for (short Counter = 0; Counter < 520; Counter += 4)
	{
		EDX = getLong(buffer, Counter);
		*(long*)&buffer[Counter] = getLong(keystream, Counter) ^ IV ^ EDX;
		IV = EDX;
	}

	if (msg[0] == 0)
		MEMCPY(buffer, msg + 2, len - 2);
	else
		MEMCPY(buffer, msg + 1, len - 1);
}


//////// Billing password encryption ////////
// Also from Coconut Emulator

void hashPassword(BYTE * in, BYTE * out)
{
	size_t StrLen = strlen((char*)in);
	BYTE Factor = simpleChecksum(in, StrLen);
	BYTE Char;

	for (DWORD L = 0; L < StrLen; L++)
	{
		Char = in[L] ^ Factor;
		Factor = (Factor ^ (Char << (Char & 3))) & 255;
		if (Char == 0)
			Char = 0xED;
		out[L] = Char;
	}

	out[L] = 0;
}


//////// Billing password decryption ////////

void inverseHash(BYTE * In, BYTE * Out, BYTE Key)
{
	size_t StrLen = strlen((char*)In);

	for (DWORD L = 0; L < StrLen; L++)
	{
		BYTE Char = In[L];
		if (Char == 0xED)
			Char = 0;
		Out[L] = Char ^ (BYTE)Key;
		Key = (Key ^ (Char << (Char & 3))) & 255;
	}
}

bool invalidPasswordCharacter(BYTE c)
{
	if ((c < 33) || (c > 126))
		return true;

	return false;
}

void decryptHashedPassword(BYTE * Password)
{
	size_t StrLen = strlen((char*)Password);

	BYTE * Buffer = new BYTE[StrLen + 1];
	Buffer[StrLen] = 0;

	// Passwords of EVEN length are very easy to crack.
	if (StrLen & 1)
	{
		// Guess and check to find one of the solutions
		for (int i = 0; i < 256; i++)
		{
			// Generate a possible solution
			inverseHash(Password, Buffer, i);

			// Compare resultant hash with given hash
			BYTE Char, Key = i;
			bool OK = true;

			for (DWORD L = 0; L < StrLen; L++)
			{
				Char = Buffer[L];

				if (invalidPasswordCharacter(Char))
				{
					OK = false;

					break;
				}

				Char ^= Key;
				Key = (Key ^ (Char << (Char & 3))) & 255;
				if (Char == 0)
					Char = 0xED;

				if (Char != Password[L])
				{
					OK = false;

					break;
				}
			}

			if (OK)
				break;
		}
	}
	else
	{
		// Generate password checksum
		inverseHash(Password, Buffer, 0);

		// Generate actual password
		inverseHash(Password, Buffer, simpleChecksum(Buffer, StrLen));
	}

	memcpy(Password, Buffer, StrLen);

	delete Buffer;
}
